﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;

namespace asprise_imaging_api
{
    /// <summary>
    /// Represents a capability.
    /// </summary>
    public class Capability
    {
        String name;
        int code;

        /** Can be: null, Integer, String, List of Int/String, CapValueEnumeration, or CapValueRange. */
        Object value;

        String errorString;
        Boolean supported = true;

        public static List<Capability> createCapList(XElement capsObject) {
            if (!capsObject.Attribute("type").Value.Equals("object"))
            {
                throw new ImagingException("Excepting an item with type object; actual: " + capsObject);
            }

            List<Capability> caps = new List<Capability>{};
            foreach (XElement xe in capsObject.Elements())
            {
                caps.Add(new Capability(xe));
            }
            return caps;
        }

        public Capability(XElement xe) {
            this.name = xe.Name.LocalName;
            this.code = TwainConstants.getCapCode(name);
            XAttribute xaType = xe.Attribute("type");
            if((xaType != null && xaType.Value.Equals("null"))) {
                this.value = null;
            } else if(xaType != null && xaType.Value.Equals("array")) {
                this.value = ScanUtils.xeElementArrayToList(xe);
            } else if(xaType != null && xaType.Value.Equals("object")) {
                if(xe.Element("ItemList") != null) {
                    this.value = new CapValueEnumeration(xe);
                } else if(xe.Element("StepSize") != null) {
                    this.value = new CapValueRange(xe);
                } else {
                    throw new ImagingException("Not recognzied: " + xe);
                }
            } else { // simple value
                string s = xe.Value;
                if(xaType != null && xaType.Value.StartsWith("bool")) {
                    this.value = s.ToLower().Equals("true");
                } else if(xaType != null && xaType.Value.Equals("number")) {
                    if(s.Contains('.')) {
                        this.value = Decimal.Parse(s);
                    } else {
                        this.value = ScanUtils.toInteger(s, -1);
                    }
                } else if(s.StartsWith("<error:")) {
                    errorString = s;
                    this.value = null;
                    this.supported = false;
                } else {
                    this.value = s; // simple value.
                }
            }
        }

        /// <summary>
        /// Returns the capability name.
        /// </summary>
        /// <returns></returns>
        public string GetName()
        {
            return name;
        }

        /// <summary>
        /// Returns the capability code.
        /// </summary>
        /// <returns></returns>
        public int GetCode()
        {
            return code;
        }

        /// <summary>
        /// Whether this capability is supported.
        /// </summary>
        /// <returns></returns>
        public bool IsSupported()
        {
            return supported;
        }
        
        /// <summary>
        /// Returns the current value of this capability.
        /// </summary>
        /// <returns></returns>
        public Object GetCurrentValue()
        {
            if (value is CapValueContainer)
            {
                return ((CapValueContainer)value).getCurrentValue();
            }
            return value;
        }

        public override string ToString()
        {
            return string.Format("{0}: {1}", name, value);
        }
    }

    // The container structure will be one of four types: TWON_ARRAY, TWON_ENUMERATION,
    // TWON_ONEVALUE, or TWON_RANGE.

    // TWON_ONEVALUE: Direct value, either Integer or String
    // TWON_ARRAY: List of Integer or String

    /// <summary>
    /// Complex capability value container.
    /// </summary>
    public abstract class CapValueContainer {

        /// <summary>
        /// The current value
        /// </summary>
        /// <returns></returns>
        abstract public Object getCurrentValue();

        /// <summary>
        /// The default value if available.
        /// </summary>
        /// <returns></returns>
        abstract public Object getDefaultValue();

        /// <summary>
        /// Check whether the attempt is valid against the enum or range. 
        /// </summary>
        /// <param name="attempt"></param>
        /// <returns></returns>
        abstract public bool validate(Object attempt);

        /// <summary>
        /// Returns the current value fall back to default.
        /// </summary>
        /// <returns></returns>
        public Object getCurrentValueFallBackToDefault() {
            Object value = getCurrentValue();
            if(value != null) {
                return value;
            }
            return getDefaultValue();
        }
    }

    /** TWON_ENUMERATION */
    /// <summary>
    /// Enumeration based capability value container.
    /// </summary>
    public class CapValueEnumeration : CapValueContainer {
        // ItemType
        // NumItems
        int CurrentIndex = -1;
        int DefaultIndex = -1;
        List<Object> ItemList;

        public CapValueEnumeration(XElement xe) {
            this.CurrentIndex = ScanUtils.xeElementValueAsInt(xe, "CurrentIndex", -1);
            this.DefaultIndex = ScanUtils.xeElementValueAsInt(xe, "DefaultIndex", -1);
            this.ItemList = ScanUtils.xeElementArrayToList(xe.Element("ItemList"));
        }

        public override Object getCurrentValue() {
            return ItemList == null || CurrentIndex < 0 ? null : ItemList.ElementAt(CurrentIndex);
        }

        public override Object getDefaultValue() {
            return ItemList == null || DefaultIndex < 0 ? null : ItemList.ElementAt(DefaultIndex);
        }

        public override bool validate(Object attempt) {
            if(attempt == null) {
                return false;
            }
            if(ScanUtils.isNumber(attempt)) { // So far, all ItemList elements are either Integer or string
                int attemptInt = ScanUtils.toInteger(attempt);
                foreach(Object item in ItemList) {
                    int number = ScanUtils.toInteger(item, -1);
                    if(number == attemptInt) {
                        return true;
                    }
                }
            } else if(attempt is string) {
                foreach(Object item in ItemList) {
                    if(((string)attempt).Equals(item.ToString())) {
                        return true;
                    }
                }
            }
            return false;
        }

        public override String ToString() {
            return "{" +
                    "CurrentIndex=" + CurrentIndex +
                    ", DefaultIndex=" + DefaultIndex +
                    ", ItemList=" + ItemList +
                    '}';
        }
    }

    /** TWON_RANGE */
    /// <summary>
    /// Range based capability value container.
    /// </summary>
    public class CapValueRange : CapValueContainer {
        // ItemType;
        Decimal MinValue;
        Decimal MaxValue;
        Decimal StepSize;
        Decimal DefaultValue;
        Decimal CurrentValue;

        public CapValueRange(XElement xe) {
            this.MinValue = ScanUtils.xeElementValueAsDecimal(xe, "MinValue", 0);
            this.MaxValue = ScanUtils.xeElementValueAsDecimal(xe, "MaxValue", 0);
            this.StepSize = ScanUtils.xeElementValueAsDecimal(xe, "StepSize", -1);
            this.DefaultValue = ScanUtils.xeElementValueAsDecimal(xe, "CurrentValue", 0);
            this.CurrentValue = ScanUtils.xeElementValueAsDecimal(xe, "DefaultValue", 0);
        }

        public override bool validate(Object attempt) {
            if(! ScanUtils.isNumber(attempt)) {
                return false;
            }

            decimal attemptNum = (attempt is decimal) ? (decimal)attempt : Decimal.Parse(attempt.ToString());

            if(attemptNum > MaxValue || attemptNum < MinValue) {
                return false;
            }

            decimal steps = (attemptNum - MinValue) / StepSize;
            decimal delta = steps - Math.Round(steps);
            if(Math.Abs(delta) < (decimal)0.02) {
                return true;
            } else {
                return false;
            }
        }

        public override Object getCurrentValue() {
            return CurrentValue;
        }

        public override Object getDefaultValue() {
            return DefaultValue;
        }

        public override String ToString() {
            return "{" +
                    "MinValue=" + MinValue +
                    ", MaxValue=" + MaxValue +
                    ", StepSize=" + StepSize +
                    ", DefaultValue=" + DefaultValue +
                    ", CurrentValue=" + CurrentValue +
                    '}';
        }

        public decimal getMin() {
            return MinValue;
        }

        public decimal getMax() {
            return MaxValue;
        }
    }
}
