﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;

#pragma warning disable 0168, 0649

namespace asprise_imaging_api
{
    /// <summary>
    /// Represents a scanner device.
    /// </summary>
    public class Source
    {
        /// <summary>
        /// Id of the scanner.
        /// </summary>
        public int Id;

        /// <summary>
        /// Twain protocol major version
        /// </summary>
        public int ProtocolMajor;

        /// <summary>
        /// Twain protocol minor version
        /// </summary>
        public int ProtocolMinor;

        /// <summary>
        /// Twain supported groups
        /// </summary>
        public int SupportedGroups;

        /// <summary>
        /// Device manufacturer.
        /// </summary>
        public String Manufacturer;

        /// <summary>
        /// Device product family.
        /// </summary>
        public String ProductFamily;

        /// <summary>
        /// Device product name - the identifier.
        /// </summary>
        public String ProductName;

        Boolean defaultSource;
        Boolean feeder;
        Boolean flatbed;

        public List<Capability> caps;

        public Source(String productName)
        {
            this.ProductName = productName;
        }

        public Source(XElement xe)
        {
            this.Id = ScanUtils.xeElementValueAsInt(xe, "Id", -1);
            this.Manufacturer = ScanUtils.xeElementValueAsString(xe, "Manufacturer");
            this.ProductFamily = ScanUtils.xeElementValueAsString(xe, "ProductFamily");
            this.ProductName = ScanUtils.xeElementValueAsString(xe, "ProductName");
            this.ProtocolMajor = ScanUtils.xeElementValueAsInt(xe, "ProtocolMajor", -1);
            this.ProtocolMinor = ScanUtils.xeElementValueAsInt(xe, "ProtocolMinor", -1);
            this.SupportedGroups = ScanUtils.xeElementValueAsInt(xe, "SupportedGroups", 0);

            this.defaultSource = "true".Equals(ScanUtils.xeElementValueAsString(xe, "default"));
            this.feeder = "true".Equals(ScanUtils.xeElementValueAsString(xe, "feeder"));
            this.flatbed = "true".Equals(ScanUtils.xeElementValueAsString(xe, "flatbed"));

            XElement xeCaps = xe.Element("caps");
            this.caps = xeCaps == null ? null : Capability.createCapList(xeCaps);
        }

        /// <summary>
        /// Whether the source is the default source.
        /// </summary>
        /// <returns></returns>
        public bool IsDefault()
        {
            return defaultSource;
        }

        /// <summary>
        /// Whether this device has a feeder
        /// </summary>
        /// <returns></returns>
        public bool HasFeeder()
        {
            return feeder;
        }

        /// <summary>
        /// Whether this device has a flatbed
        /// </summary>
        /// <returns></returns>
        public bool HasFlatbed()
        {
            return flatbed;
        }

        /// <summary>
        /// Gets all capabilities
        /// </summary>
        /// <returns></returns>
        public List<Capability> GetCaps()
        {
            return caps;
        }

        /// <summary>
        /// Gets the capability of the given name.
        /// </summary>
        /// <param name="capName">The cap name</param>
        /// <returns>the capability of the given name or null if not found.</returns>
        public Capability GetCap(String capName)
        {
            if (capName == null || caps == null)
            {
                return null;
            }

            foreach(Capability cap in caps) {
                if(ScanUtils.equalsIgnoreCase(capName, cap.GetName())) {
                    return cap;
                }
            }
            return null;
        }

        /// <summary>
        /// Gets the capability of the given code.
        /// </summary>
        /// <param name="capCode"></param>
        /// <returns>the capability of the given code or null if not found.</returns>
        public Capability GetCap(int capCode)
        {
            return GetCap(TwainConstants.getCapName(capCode));
        }

        /// <summary>
        /// Returns the source object with given name or null if not found.
        /// </summary>
        /// <param name="sources"></param>
        /// <param name="name"></param>
        /// <returns></returns>
        public static Source GetSourceByName(List<Source> sources, String name) {
            if(sources == null || name == null) {
                return null;
            }

            foreach(Source source in sources) {
                if(source != null && name.Equals(source.ProductName)) {
                    return source;
                }
            }
            return null;
        }

        /// <summary>
        /// Returns the first source marked as default in the first or null if not found.
        /// </summary>
        /// <param name="sources"></param>
        /// <returns></returns>
        public static Source GetDefaultSource(List<Source> sources) {
            if(sources == null) {
                return null;
            }

            foreach(Source source in sources) {
                if(source != null && source.defaultSource) {
                    return source;
                }
            }
            return null;
        }

        /// <summary>
        /// Creates a list of Sources from the given JSON.
        /// </summary>
        /// <param name="xeArray"></param>
        /// <returns></returns>
        public static List<Source> createSourceList(XElement xeArray) {
            if (!xeArray.Attribute("type").Value.Equals("array"))
            {
                throw new ImagingException("Expecting type=array, actual: " + xeArray);
            }

            List<Source> sources = new List<Source> { };
            // Ref: https://mutelight.org/using-the-little-known-built-in-net-json-parser
            foreach (XElement item in xeArray.Elements())
            {
                Source src = new Source(item);
                sources.Add(src);
            }
            return sources;
        }

        /// <summary>
        /// Reads XML attribute and returns its int value.
        /// </summary>
        /// <param name="reader"></param>
        /// <param name="attrName"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public static int xmlAttrAsInt(XmlReader reader, string attrName, int defaultValue) {
            string value = reader.GetAttribute(attrName);
            if (value == null || value.Trim().Length == 0)
            { 
                return defaultValue;
            }
            try
            {
                return Int32.Parse(value.Trim());
            }
            catch (FormatException fe)
            {
                return defaultValue;
            }
        }


        public override string ToString()
        {
            return ProductName;
        }
    }
}
